/***************************************************************************************************
PhotoFinish.c
History: Loosely based on the MicroMatic project but cannot load commands to be executed later.
Reference: "PhotoFinish #1" MAC: 00-90-C2-CC-55-50.
***************************************************************************************************/

//STILL TO DO
//===========================
// * add loss of incoming heartbeat to disconnect network and restart it.
// * at start, enable retry connect to network if not plugged in at boot.
// * when waiting for host or ready, if lose e'net, reconnect.


#define VERSION  "1.2.4"								 //appears on the monitor and display, max 11 chars.
//vvvvvvvvvvvvvvvvvvvvvvvvvv COMMENT OUT FOLLOWING LINE vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
//#define DO_DEBUG  	  //normally commented out, otherwise will allow debugging through serial port.
#class auto
//#use rcm3200.lib													 //Rabbit's housekeeping (inc IP) routines.
#use "Unit Identification.h"											 //the unique name and IP for this unit.
#use "Rabbit Command.lib"																//command parsing functions.
#use "Rabbit Hardware.lib"												 //hardware drivers for DAC and display.
#use "Rabbit Monitor.lib"	 														  //monitor (Warren) functions.
#use "Rabbit Help.lib"									 //all of the project's function helps (for Ctrl-H).

#ifdef DO_DEBUG
   #warnt "FINAL RELEASE WARNING: Running with debugging turned on!"
#else
	#nodebug								// disables insertion of breakpoints for debugging, this file only.
#endif
#define MONITOR_IP 		 	 UNIT_MONITOR_ADDRESS					 //as set in 'Unit Identification.h'.
#define _PRIMARY_NETMASK	 "255.255.255.0"	 //Reference: Rabbit RCM3200 #1 MAC: 00-90-C2-CC-55-50.
#define MY_PORT 				37829									 //all Rabbits use this port for the panel.
//do not uncomment next two lines unless you want to slog through the debug info.
//#define DCRTCP_DEBUG													// display tcp/ip debugging information.
//#define DCRTCP_VERBOSE											// allow stepping through tcp/ip functions.

//project specific constants
#define DAC_DEFAULT		 1500												 //default DAC values, both channels.
#define HEARTBEATPERIOD  2000														//time between heartbeats (mS).
#define SINGLES_PERIOD   1100									 //time between reporting singles counts (mS).
#define MONITOR_PORT		 6595										 // Port for UDP socket server for monitor.
#define MTU_R				 1400	 		 /*maximum TCP packet size for coincidence reports sent to host.
								       Was MTU_E-50. Changed for critical process. Now MTU_R = MTU - 50 - 7.
						 Note that host probably does not reassemble: each packet is treated seperately.*/

//next 11 lines are required by system
#define TCPCONFIG 			1	 									 //See LIB\TCP_CONFIG.LIB for instructions.
#define MTU_E			 		1460 //Need to change MTU_R above!!
#define BUFF_SIZE	 			(MTU_E-40)	 //must be less than (ETH_MTU - (IP Datagram + TCP Segment)).
#define UDP_BUF_SIZE 		10000
#define INCOMING_IP			0							  								// 0 accept all connections.
#define INCOMING_PORT		0							  										// 0 accept all ports.
//#define USE_IF_CALLBACK							  					  // Call the interface TcpStatus callback.
#use "dcrtcp.lib"
#define STDIO_DISABLE_FLOATS			// Disables floating point support, to save space program space.
#define FAST_INTERRUPT 0
#memmap xmem												 // Sets program functions to run in xternal memory.

/* function prototypes */
void CoincidenceCheck(void);								 //if running, check for and report coincidences.
void SinglesReport(void);					  //get singles count from FPGA, send to host, clear in FPGA.
void GetCoinData (void);
void Pause (int mS);

/* Global Variables */
udp_Socket sock;																					  //the UDP/IP socket.
int TcpBuffPtr;															  //to parse through incomming commands.
int TcpStatus;
unsigned long TimedCoinCount;		 							//counter for timed coincidences when running.
char CoinRpt[MTU_R+10];		//buffer for coincidence reports, sent out every second, or less if full.
unsigned int  CoinRptLen;		  //because report may contain 0x00, can't use strlen() to get length.
int  Calibrating;			//flag to say when calibrating, so as not to send multiple monitor messages.
unsigned long HeartbeatLastOut, LastDisplayUpdate, NextSinglesReport;
char MsgFifo[1000];						  			  //buffer for saving unsent messages, to be sent soon.
char Msg[71];										 //buffers for formatting messages for monitor or display.
int TimeStampEn;																	//Enable the time stamp from FPGA.
int CountTime;


/*********************************** Function main() **********************************************/
void main() {			/* Concerned with handling the TCP/IP connection to the panel and
								overseeing all other functions including scans, monitor, TcpStatus, etc. */
	int TcpLen;
	unsigned long HeartbeatLastIn;
	unsigned char TcpBuff[BUFF_SIZE];

	Boot();							 											 				 /*INITIAL HOUSEKEEPING*/
	HeartbeatLastIn = 0;
	NextSinglesReport = MS_TIMER;													//actually, not used, we think.
	LastDisplayUpdate = 0;
	Calibrating = 0;
   TimeStampEn = 0;																		//Disable the time stamping.
   CountTime 	= 0;   																  //Count forever once started.

	strcpy(Msg, "FPGA: ");																			//Get Fpga Version.
	SetFpgaAddress(13);
	utoa(RdPortI(PGDR)&0x0F    , Msg+6);					  												  //major.
	strcat(Msg, ".");
	utoa((RdPortI(PFDR)&0xF0)/16, Msg+8);																	  //minor.
	strcat(Msg, ".");
	utoa((RdPortI(PFDR)&0x0F)   , Msg+10);																  //revision.
	strcpy(TcpBuff, "Rev: ");  strcat(TcpBuff, VERSION);
	DisplayLumex2x16(TcpBuff, Msg);
	LastDisplayUpdate = MS_TIMER;
	Pause (2000);

	DisplayLumex2x16(UNIT_NAME, "Network???");
	sock_init();	 																				//INITIATE UDP SOCKET.
	if (!udp_open(&sock, MY_PORT, 0, MY_PORT, NULL)); 		  //OPEN A UDP SOCKET TO THE HOST.
  	else  {
		while(!sock_established(&sock))
  			if (!tcp_tick(&sock))  break;
  	}
	Pause (1000);

	MonitorMain();														 // initiate the UDP socket to the monitor.
	sprintf(Msg, "%s is starting, Rev: %s.\n", UNIT_NAME, VERSION);  MonitorSendMsg(Msg);
	#ifdef DO_DEBUG
		sprintf(Msg, "Warning: %s is in debug mode.\n", UNIT_NAME);  MonitorSendMsg(Msg);
	#endif

/*************************** Main Loop ****************************/
	while(1)  {																			// MAIN LOOP, RUN HERE FOREVER.
		tcp_tick (NULL);														 //required periodic UDP maintenance.

		CoincidenceCheck();

		TcpLen = udp_recv(&sock, TcpBuff, BUFF_SIZE);						  //read incoming data, if any.
		if (TcpLen > 0)  {															//Incoming Command, Process It.
			TcpBuffPtr = 0;
			while(TcpBuffPtr < TcpLen)  {									  //Extract commands from TCP buffer.
				if ((TcpBuff + TcpBuffPtr)[0] == (char)'H')  { 										 //heartbeat!
					udp_send(&sock, "H", 1);													 //echo the heartbeat.
					TcpBuffPtr++;
				}
				else	CommandHandler(TcpBuff + TcpBuffPtr, TcpLen);						  //a real command.
			}
		}
	}
}


/************************************ CoincidenceCheck ********************************************/
void CoincidenceCheck(void)  {	/* If running, check for and report coincidences.  Also, check,
												report and reset singles counters. */
	unsigned long CoinRepeat;

	if (BitRdPortI(PGDR, 7))  {																				//RUNNING?
		CoinRepeat = MS_TIMER + 200;
		while (MS_TIMER < CoinRepeat)  {							//KEEP CHECKING FOR COINCIDENCES FOR 100mS.
			if (NextSinglesReport <= MS_TIMER)	SinglesReport();//handle singles counts.

        /*if (BitRdPortI(PEDR, 1))  {						 									//HAVE COINCIDENCE.
			if (TimeStampEn)	{
					if (!CoinRptLen)  {		 							//make coincidence report packet header.
						CoinRpt[0] = 'P';
						CoinRptLen = 2;
					}*/

        	//if (TimeStampEn)
         #asm
         ld hl, (TimeStampEn)
			bool hl
			jp z, NoCoinStp

	         //if (BitRdPortI(PEDR, 1))
         	ioi ld a, (0x0070)
				and 0x02
				jp z, NoCoinStp

               //@@BitWrPortI(PGDR, &PGDRShadow, 1, 6); 		  //Rabbit start reading Data from FPGA.
               IOI LD A, (PGDR)
               SET 6, A					;Set RabbitReadingData high.
               IOI LD (PGDR), A

               //if (!CoinRptLen)
					ld hl, (CoinRptLen)
					bool hl
               jp nz, RptLenNonZero

						//CoinRpt[0] = 'P';
					ld hl, CoinRpt
					ld (hl), 0x50

						//CoinRptLen = 2;
					ld hl, 0x0002
					ld (CoinRptLen), hl

               //CoinRpt[CoinRptLen++] = 'R';		 							  //coincidence report flag.
RptLenNonZero: LD HL, CoinRpt     		;load CoinRpt
               LD BC, (CoinRptLen) 	 	;load offset
               ADD HL, BC 					;add offset
               LD A, 82						;load "R"
               LD (HL), A           	;write "R" to address, byte 1/7
               ;LD HL, (CoinRptLen)  	;load CoinRptLen
               ;INC HL               	;inc CoinRptLen
					;LD (CoinRptLen), HL  	;save it

//					SetFpgaAddress(0);
					IOI LD A, (PEDR)		;read in PEDR
               AND 0x0F          	;bit 4765 are 0000
               SET 3, A
               LD  D, A					;save PEDR to D, CS high, add 0
               IOI LD (PEDR), A		;write to PEDR, CS high, add 0
					AND 0x07
			  		IOI LD (PEDR), A		;write to PEDR, CS low, add 0

			 		//CoinRpt[CoinRptLen++] = RdPortI(PGDR) & 0x0F;	  				  		//ADC A: MS nibble.
               INC HL
               IOI LD A, (PGDR)		;Read PortG
					AND 0x0F             ;Mask with 0x0F
               LD (HL), A           ;write PortG to address, byte 2/7

               //CoinRpt[CoinRptLen++] = RdPortI(PFDR);	 	 										//LS byte.
               INC HL
               IOI LD A, (PFDR)		;Read PortF
               LD (HL), A           ;write PortF to address, byte 3/7

//					BitWrPortI(PEDR, &PEDRShadow, 1, 5);		 	  					 //ADC "B" in FPGA reg 1.
               LD  A, D					;read in PEDR, add0
               SET 5, A             ;add 1
					IOI LD (PEDR), A		;write 1 to PEDR, CS high
               RES 3, A
  					IOI LD (PEDR), A		;CS low

		  			//CoinRpt[CoinRptLen++] = RdPortI(PGDR) & 0x0F;							//ADC B: MS nibble.
               INC HL
               IOI LD A, (PGDR)		;Read PortG
					AND 0x0F             ;Mask with 0x0F
               LD (HL), A           ;write PortG to address, byte 4/7

			  		//CoinRpt[CoinRptLen++] = RdPortI(PFDR);
               INC HL
               IOI LD A, (PFDR)		;Read PortF
               LD (HL), A           ;write PortF to address, byte 5/7


//			  		BitWrPortI(PEDR, &PEDRShadow, 0, 5);	  		  	//offset and overrange in FPGA reg 2.
// 		  		BitWrPortI(PEDR, &PEDRShadow, 1, 6);
               LD  A, D					;read in PEDR, add 0
			  		SET 6, A					;bit 4765 are 0010, add 2
			  		IOI LD (PEDR), A		;write to PEDR, CS high
               RES 3, A
  					IOI LD (PEDR), A		;CS low

               //CoinRpt[CoinRptLen++] = RdPortI(PFDR);				 //offset counter, 10nS per count.
               INC HL
               IOI LD A, (PFDR)		;Read PortF
               LD (HL), A           ;write PortF to address, byte 6/7

			 		//CoinRpt[CoinRptLen++] = RdPortI(PGDR) & 0x07;	 				  //overranges, PD order.
               INC HL
               IOI LD A, (PGDR)		;Read PortG
					AND 0x0F             ;Mask with 0x0F
               LD (HL), A           ;write PortG to address, byte 7/7

               LD HL, (CoinRptLen)  ;load CoinRptLen
               LD BC, 0x0007
               ADD HL, BC           ;increase CoinRptLen by 7
               LD (CoinRptLen), HL  ;save it

               /*if (CoinRptLen > MTU_R)  {			  	  //FULL PACKET, SEND COINCIDENCE REPORT TO HOST.
			  			CoinRpt[1] = 200;//(CoinRptLen - 2) / 7;//number of coincidence reports in packet.
			  			udp_send(&sock, CoinRpt, CoinRptLen);									 //send the report.
                  tcp_tick(NULL); 			    											// socket maintenance.
			  			CoinRptLen = 0;								  								//clear the report.
              	}*/

					ld de, MTU_R				;if (CoinRptLen > MTU_R)
					or a
					sbc hl, de
					jp c, NoPkt
					ld hl, CoinRpt				;CoinRpt[1] = 200;
					inc hl
					ld (hl), 200
               								;udp_send(&sock, CoinRpt, CoinRptLen)
               ld bc, 0                ;udp_send() is  assembled with udp_xsendto().
               push bc						;remport = 0, word
               push bc                 ;remip   = 0, longword
               push bc
               ld hl, (CoinRptLen)
               push hl
               ld hl, CoinRpt
					push hl
               lcall paddr
					add sp, 2
					push bc                 ;paddr return value LS?
					push de                 ;paddr return value MS?
	  				ld hl, sock
					push hl
               lcall _rs_udp_xsendto
					add sp, 14
					ld hl, 0x00   				;tcp_tick(NULL);
					push hl
					lcall _rs_tcp_tick
					add sp, 2
					ld hl, 0x0000 				;CoinRptLen = 0;
					ld (CoinRptLen), hl

               //TimedCoinCount++;
NoPkt:         LD HL, (TimedCoinCount)
               LD BC, 0x01
               ADD HL, BC
               LD (TimedCoinCount), HL
               JR NC, noc1					;jump if carrier bit C is 0
               LD HL, TimedCoinCount
               INC HL
               INC HL
               LD A, 0x00
               ADC A, (HL)
               LD (HL), A
               JR NC, noc1
               INC HL
					INC (HL)

               //@@BitWrPortI(PGDR, &PGDRShadow, 0, 6); 	  //Rabbit finished reading Data from FPGA.

noc1: 		  	IOI LD A, (PGDR)		;read in PGDR
			  		RES 6, A					;set RabbitReadingData low
				  	IOI LD (PGDR), A		;write to PGDR
NoCoinStp: 	   NOP
               #endasm
		}
 	}
}


/************************************ SinglesReport ***********************************************/
void SinglesReport(void)  {		/* Get singles counts from FPGA, send to host, reset counters.
			   Singles are ADC counts passing threshold per channel, whether or not a coincidence.
			   Enter in normal coincidence detection mode (PE[3:4] = 0x00).
			   Note that the FPGA will stop counting singles at 999,999. */

	unsigned long HitsA, HitsB, CoinCount;
   unsigned long rWaitTime,fWaitTime;
	char HitsRpt[16], *Hit;	  													 //buffer for hits report to host.
   char DisLin1[20],DisLin2[20];

   while(MS_TIMER  < NextSinglesReport + 5);			  //wait for 5ms here, to make sure a precise 1s.
   BitWrPortI(PGDR, &PGDRShadow, 0, 7);	 				  //stop running so can read singles counters.
	NextSinglesReport = NextSinglesReport + SINGLES_PERIOD;					 //set time for next report.
   CountTime--;
   rWaitTime = MS_TIMER + SINGLES_PERIOD;							 //Rabbit WaitTime, to get a precise 1s.
  	#ifdef DO_DEBUG
		printf("Start: %lu - ", MS_TIMER);
	#endif

                                                                                    //READ COUNTERS.
//	SetFpgaAddress(5);
	#asm
	IOI LD A, (PEDR)		;read in PEDR
	SET 5, A					;FPGA reg 5
	RES 6, A
	SET 7, A
	RES 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR
	RES 3, A
	IOI LD (PEDR), A		;write to PEDR, CS low
	#endasm
	HitsA  =  (unsigned long)RdPortI(PFDR);
	HitsA += ((unsigned long)RdPortI(PGDR) & 0x0F) * 256;
/*	BitWrPortI(PEDR, &PEDRShadow, 0, 3);	//CS low
*/

/*
	BitWrPortI(PEDR, &PEDRShadow, 0, 5);				  												//FPGA reg 6.
	BitWrPortI(PEDR, &PEDRShadow, 1, 6);
	BitWrPortI(PEDR, &PEDRShadow, 1, 7);
	BitWrPortI(PEDR, &PEDRShadow, 0, 4);
	BitWrPortI(PEDR, &PEDRShadow, 1, 3);	//CS high
*/
	#asm
	IOI LD A, (PEDR)		;read in PEDR
	RES 5, A					;FPGA reg 6
	SET 6, A
	SET 7, A
	RES 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR.
	RES 3, A
	IOI LD (PEDR), A		;write o PEDR, CS low.
	#endasm
	HitsB  =  (unsigned long)RdPortI(PFDR);
	HitsB += ((unsigned long)RdPortI(PGDR) & 0x0F) * 256;
/*	BitWrPortI(PEDR, &PEDRShadow, 0, 3);	//CS low
*/

/*
	BitWrPortI(PEDR, &PEDRShadow, 1, 5);			  													//FPGA reg 7.
	BitWrPortI(PEDR, &PEDRShadow, 1, 6);
	BitWrPortI(PEDR, &PEDRShadow, 1, 7);
	BitWrPortI(PEDR, &PEDRShadow, 0, 4);
	BitWrPortI(PEDR, &PEDRShadow, 1, 3);	//CS high
*/
	#asm
	IOI LD A, (PEDR)		;read in PEDR
	SET 5, A					;FPGA reg 7
	SET 6, A
	SET 7, A
	RES 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR.
	RES 3, A
	IOI LD (PEDR), A		;write to PEDR, CS low.
	#endasm
	HitsA += ((unsigned long)RdPortI(PFDR)) * 4096;	//2^12
	HitsA += ((unsigned long)RdPortI(PGDR)&0x0F) * 1048576;	//2^20
/*	BitWrPortI(PEDR, &PEDRShadow, 0, 3);	//CS low
*/

/*
	BitWrPortI(PEDR, &PEDRShadow, 0, 5);			  													//FPGA reg 8.
	BitWrPortI(PEDR, &PEDRShadow, 0, 6);
	BitWrPortI(PEDR, &PEDRShadow, 0, 7);
	BitWrPortI(PEDR, &PEDRShadow, 1, 4);
	BitWrPortI(PEDR, &PEDRShadow, 1, 3);	//CS high
*/
	#asm
	IOI LD A, (PEDR)		;read in PEDR
	RES 5, A					;FPGA reg 8
	RES 6, A
	RES 7, A
	SET 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR. FPGA reg 8, CS high.
	RES 3, A
	IOI LD (PEDR), A		;write to PEDR, CS low.
	#endasm
	HitsB += ((unsigned long)RdPortI(PFDR)) * 4096;	//2^12
	HitsB += ((unsigned long)RdPortI(PGDR)&0x0F) * 1048576;	//2^20
/*	BitWrPortI(PEDR, &PEDRShadow, 0, 3);	//CS low
*/
   //	SetFpgaAddress(14);
	#asm
	IOI LD A, (PEDR)		;read in PEDR
	RES 5, A					;FPGA reg 14
	SET 6, A
	SET 7, A
	SET 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR
	RES 3, A
	IOI LD (PEDR), A		;write to PEDR, CS low
	#endasm
	CoinCount  =  (unsigned long)RdPortI(PFDR);
	CoinCount += ((unsigned long)RdPortI(PGDR) & 0x0F) * 256;

   //	SetFpgaAddress(15);
   #asm
	IOI LD A, (PEDR)		;read in PEDR
	SET 5, A					;FPGA reg 15
	SET 6, A
	SET 7, A
	SET 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR.
	RES 3, A
	IOI LD (PEDR), A		;write to PEDR, CS low.
	#endasm
	CoinCount += ((unsigned long)RdPortI(PFDR)) * 4096;	//2^12
	CoinCount += ((unsigned long)RdPortI(PGDR)&0x0F) * 1048576;	//2^20

   //	SetFpgaAddress(3);                               //FPGA WAIT TIME FOR RABBIT'S ACKNOWLEGEMENT.
	#asm
	IOI LD A, (PEDR)		;read in PEDR
	SET 5, A					;FPGA reg 3
	SET 6, A
	RES 7, A
	RES 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR
	RES 3, A
	IOI LD (PEDR), A		;write to PEDR, CS low
	#endasm
	fWaitTime  =  (unsigned long)RdPortI(PFDR);
	fWaitTime += ((unsigned long)RdPortI(PGDR) & 0x0F) * 256;

   //	SetFpgaAddress(4);
   #asm
	IOI LD A, (PEDR)		;read in PEDR
	RES 5, A					;FPGA reg 4
	RES 6, A
	SET 7, A
	RES 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR.
	RES 3, A
	IOI LD (PEDR), A		;write to PEDR, CS low.
	#endasm
	fWaitTime += ((unsigned long)RdPortI(PFDR)) * 4096;	//2^12
	fWaitTime += ((unsigned long)RdPortI(PGDR)&0x0F) * 1048576;	//2^20

   Hit = (char*) &HitsA;																//CHANGE INTO REPORT FORMAT.
	HitsRpt[1] = Hit[2];
	HitsRpt[2] = Hit[1];
	HitsRpt[3] = Hit[0];
	Hit = (char*) &HitsB;
	HitsRpt[4] = Hit[2];
	HitsRpt[5] = Hit[1];
	HitsRpt[6] = Hit[0];
   Hit = (char*) &CoinCount;
	HitsRpt[7] = Hit[2];
	HitsRpt[8] = Hit[1];
  	HitsRpt[9] = Hit[0];
   Hit = (char*) &TimedCoinCount;
 	HitsRpt[10] = Hit[2];
	HitsRpt[11] = Hit[1];
  	HitsRpt[12] = Hit[0];
   Hit = (char*) &fWaitTime;
 	HitsRpt[13] = Hit[2];
	HitsRpt[14] = Hit[1];
  	HitsRpt[15] = Hit[0];

   HitsRpt[0] = 'S';																  				 	 //SEND THE REPORT.
   udp_send(&sock, HitsRpt, 16);										 		 				 	 //send the report.

   sprintf(DisLin1,"A%lu B%lu", HitsA, HitsB);							  				 //format for display.
	DisLin1[19] = 0;																				//ensure not too long.
	sprintf(DisLin2, "AB%lu T%lu", CoinCount, TimedCoinCount);
	DisplayLumex2x16(DisLin1, DisLin2);			  			 //display singles rates and coincidence rate.
	#ifdef DO_DEBUG
		printf("A: %lu B: %lu AB: %lu Timed: %lu Wait: %lu", \
      HitsA, HitsB, CoinCount, TimedCoinCount, fWaitTime);
	#endif

   if	(CoinRptLen > 0 )  {									  						  //Send Coincidence Report(s).
		CoinRpt[1] = (CoinRptLen - 2) / 7;				 		 //number of coincidence reports in packet.
      udp_send(&sock, CoinRpt, CoinRptLen); 									  				 //send the report.
      CoinRptLen = 0;								  									  			//clear the report.
	}
   HeartbeatLastOut = MS_TIMER;				  					 	 //no need to send a seperate heartbeat.
	TimedCoinCount = 0;
   if (!CountTime) {
   	udp_send(&sock, "D", 1);
      DisplayLumex2x16(UNIT_NAME, "Ready");
   }
   tcp_tick(NULL);
   while(MS_TIMER  < rWaitTime - 1000);           						//A constant time interval between
   																						  //resetting the FPGA counter.
/*
	SetFpgaAddress(9);                                          	//RESET COUNTERS & RESUME RUNNING.
	BitWrPortI(PGDR, &PGDRShadow, 1, 7);	 												  //re-enable running.
*/
   #asm
	IOI LD A, (PEDR)		;read in PEDR
	SET 5, A					;FPGA reg 9
	RES 6, A
	RES 7, A
	SET 4, A
	SET 3, A					;CS high
	IOI LD (PEDR), A		;write to PEDR.
	RES 3, A
	IOI LD (PEDR), A		;write to PEDR, CS low.

   LD  HL, (CountTime)
   BOOL HL
   JR Z, StopRun

	IOI LD A, (PGDR)		;read in PGDR
	SET 7, A
	IOI LD (PGDR), A		;write to PGDR, re-enable running.
StopRun:
   NOP
   #endasm
 	#ifdef DO_DEBUG
		printf(" - End: %lu\n", MS_TIMER);
	#endif
}


/****************************************** Pause *************************************************/
void Pause (int mS)  {		//Wait here for given mS.  Description in "PC AWG Hardware.c".
	unsigned long Paws;

	Paws = MS_TIMER;
	while (Paws + mS > MS_TIMER);
}


/**************************************************************************************************/


