/***************************************************************************************************
PhotoFinish.c
History: Loosely based on the MicroMatic project but cannot load commands to be executed later.
Reference: "PhotoFinish #1" MAC: 00-90-C2-CC-55-50.
***************************************************************************************************/

//STILL TO DO
//===========================
// * BUG: may lockup after a few dozen coincidences reported, problem is in FPGA.
// * add loss of incoming heartbeat to disconnect network and restart it.
// * at start, enable retry connect to network if not plugged in at boot.
// * when waiting for host or ready, if lose e'net, reconnect.


#define VERSION  "1.1.0"								 //appears on the monitor and display, max 11 chars.
//vvvvvvvvvvvvvvvvvvvvvvvvvv COMMENT OUT FOLLOWING LINE vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
//#define DO_DEBUG  	  //normally commented out, otherwise will allow debugging through serial port.
#class auto
#use rcm3200.lib														 //Rabbit's housekeeping (inc IP) routines.
#use "Unit Identification.h"											 //the unique name and IP for this unit.
#use "Rabbit Command.lib"																//command parsing functions.
#use "Rabbit Hardware.lib"												 //hardware drivers for DAC and display.
#use "Rabbit Monitor.lib"	 														  //monitor (Warren) functions.
#use "Rabbit Help.lib"									 //all of the project's function helps (for Ctrl-H).

#ifdef DO_DEBUG
   #warnt "FINAL RELEASE WARNING: Running with debugging turned on!"
#else
	#nodebug								// disables insertion of breakpoints for debugging, this file only.
#endif
#define _PRIMARY_STATIC_IP  UNIT_IP_ADDRESS							 //as set in 'Unit Identification.h'.
#define MONITOR_IP 		 	 UNIT_MONITOR_ADDRESS					 //as set in 'Unit Identification.h'.
#define _PRIMARY_NETMASK	 "255.255.255.0"	 //Reference: Rabbit RCM3200 #1 MAC: 00-90-C2-CC-55-50.
#define MY_PORT 				37829									 //all Rabbits use this port for the panel.
//do not uncomment next two lines unless you want to slog through the debug info.
//#define DCRTCP_DEBUG													// display tcp/ip debugging information.
//#define DCRTCP_VERBOSE											// allow stepping through tcp/ip functions.

//project specific constants
#define DAC_DEFAULT		 1500												 //default DAC values, both channels.
#define HEARTBEATPERIOD  2000														//time between heartbeats (mS).
#define SINGLES_PERIOD   1000									 //time between reporting singles counts (mS).
#define MONITOR_PORT		 6595										 // Port for UDP socket server for monitor.
#define MTU_R				(MTU_E-50)	 /*maximum TCP packet size for coincidence reports sent to host.
						 Note that host probably does not reassemble: each packet is treated seperately.*/

//next 11 lines are required by system
#define TCPCONFIG 			1	 									 //See LIB\TCP_CONFIG.LIB for instructions.
#define MTU_E			 		1000
#define BUFF_SIZE	 			(MTU_E-40)	 //must be less than (ETH_MTU - (IP Datagram + TCP Segment)).
#define TCP_BUF_SIZE 		((MTU_E-40)*4)
#define INCOMING_IP			0							  								// 0 accept all connections.
#define INCOMING_PORT		0							  										// 0 accept all ports.
#define USE_IF_CALLBACK							  					  // Call the interface TcpStatus callback.
#use "dcrtcp.lib"
#define STDIO_DISABLE_FLOATS			// Disables floating point support, to save space program space.
#define FAST_INTERRUPT 0
#memmap xmem												 // Sets program functions to run in xternal memory.

/* function prototypes */
void CoincidenceCheck(void);								 //if running, check for and report coincidences.
void SinglesReport(void);					  //get singles count from FPGA, send to host, clear in FPGA.

/* Global Variables */
tcp_Socket sock;																					  //the TCP/IP socket.
int TcpBuffPtr;															  //to parse through incomming commands.
long CoinCount, CoinCountLast;		//counter for number of coincidences when running, and previous.
char CoinRpt[MTU_R+10];		//buffer for coincidence reports, sent out every second, or less if full.
int  CoinRptLen;					  //because report may contain 0x00, can't use strlen() to get length.
int  Calibrating;			//flag to say when calibrating, so as not to send multiple monitor messages.
unsigned long HeartbeatLastOut, LastDisplayUpdate, LastSinglesReport;
char MsgFifo[1000];						  			  //buffer for saving unsent messages, to be sent soon.
char Msg[71], SinglesMsg[20];					 //buffers for formatting messages for monitor or display.


/*********************************** Function main() **********************************************/
void main() {			/* Concerned with handling the TCP/IP connection to the panel and
								overseeing all other functions including scans, monitor, TcpStatus, etc. */
	int MainState, TcpStatus, TcpLen, LastState;
	unsigned long HeartbeatLastIn;
	unsigned char TcpBuff[BUFF_SIZE];

	Boot();							 											 				 /*INITIAL HOUSEKEEPING*/
	MainState = 0;
	LastState = 99;													  //force first pass to force state change.
	HeartbeatLastIn = 0;
	LastSinglesReport = 0;
	LastDisplayUpdate = 0;
	Calibrating = 0;

	strcpy(Msg, "FPGA: unknown version");														//Get Fpga Version.
/*	strcpy(Msg, "FPGA: 0");																			//Get Fpga Version.
	BitWrPortI(PEDR, &PEDRShadow, 0, 5);												 //set address LSB low...
	BitWrPortI(PEDR, &PEDRShadow, 1, 5);															 //...then high.
	BitWrPortI(PEDR, &PEDRShadow, 1, 6);
	BitWrPortI(PEDR, &PEDRShadow, 1, 7);															  //address MSB.
	utoa(RdPortI(PGDR)&0x0F    , Msg+7);					  												  //major.
	strcat(Msg, ".0");
	utoa((RdPortI(PFDR)&0xF0)/16, Msg+10);																	  //minor.
	strcat(Msg, ".0");
	utoa((RdPortI(PFDR)&0x0F)   , Msg+13);																  //revision.
*/	strcpy(TcpBuff, "Rev: ");  strcat(TcpBuff, VERSION);
	DisplayLumex2x16(TcpBuff, Msg);

	DisplayLumex2x16(UNIT_NAME, "Network???");
	#ifdef DO_DEBUG					 				  							  // Show commands when debugging.
		printf("Starting, looking for network connection...\n");
	#endif
	sock_init();	 																  // INITIATE TCP SOCKET TO PANEL.
	HeartbeatLastOut = MS_TIMER;  while (MS_TIMER < HeartbeatLastOut + 1000);		  //pause a second.
	while (ifpending(IF_DEFAULT) == IF_COMING_UP)  tcp_tick(NULL);	  					 //wait to come up.

	MonitorMain();														 // initiate the UDP socket to the monitor.
	sprintf(Msg, "%s is starting, Rev: %s.\n", UNIT_NAME, VERSION);  MonitorSendMsg(Msg);
	#ifdef DO_DEBUG
		sprintf(Msg, "Warning: %s is in debug mode.\n", UNIT_NAME);  MonitorSendMsg(Msg);
	#endif


	while(1)  {																			// MAIN LOOP, RUN HERE FOREVER.
		TcpStatus = tcp_tick(&sock);															// socket maintenance.

		if ((MS_TIMER > HeartbeatLastOut + HEARTBEATPERIOD) & (sock_established(&sock)))  {	//Send HB.
			sock_fastwrite(&sock, "H", 1);
			HeartbeatLastOut = MS_TIMER;
		}

		if( MainState != LastState)  {														 //DISPLAY THE STATUS.
			switch (MainState)  {
				case 1:	DisplayLumex2x16(UNIT_NAME, "Waiting for host");  break;
				case 2:	DisplayLumex2x16(UNIT_NAME, "Ready");  break;
				default:  ;
			}
			LastState = MainState;
//			#ifdef DO_DEBUG
//				printf("Panel TCP command state = %d.\n", MainState);
//			#endif
		}

		switch(MainState)  {
			case 0:																	/* OPEN TCP CONNECTION TO PANEL */
				if(tcp_listen(&sock, MY_PORT, INCOMING_IP, INCOMING_PORT, NULL, 0))			// connected!
					MainState = 1;
				break;


 			case 1:															/* WAITING FOR CONNECTION TO PANEL... */
 				if(sock_established(&sock))  MainState = 2;		  							  // Now Connected.
				BitWrPortI(PGDR, &PGDRShadow, 0, 7);									 //PG7, kill RunEn[able].
				BitWrPortI(PEDR, &PEDRShadow, 0, 4);								 //PE4, set calibration off.
				BitWrPortI(PEDR, &PEDRShadow, 0, 3);								 //PE3, set calibration low.
				break;


			case 2:		 								/* CONNECTED TO PANEL, WAIT FOR AND EXECUTE COMMANDS */
 				if(!sock_established(&sock))  MainState = 3;								//No longer connected.
				TcpLen = sock_fastread(&sock, TcpBuff, BUFF_SIZE);			  //read incoming data, if any.
				CoincidenceCheck();

				if (TcpLen > 0)  {													//Incoming Command, Process It.
					TcpBuffPtr = 0;
					while(TcpBuffPtr < TcpLen)  {							  //Extract commands from TCP buffer.
						if ((TcpBuff + TcpBuffPtr)[0] == (char)'H')  TcpBuffPtr++;  		 //heartbeat in.
						else	CommandHandler(TcpBuff + TcpBuffPtr, TcpLen);							//command.
					}
				}
				break;


			case 3:																					 /* CLOSE THE SOCKET*/
				sock_close(&sock); 																  // close the socket.
				while(sock_alive(&sock))  sock_abort(&sock);				// wait for socket to fully close.
				MainState = 0;
				break;
		}
	}
}


/************************************ CoincidenceCheck ********************************************/
void CoincidenceCheck(void)  {	/* If running, check for and report coincidences.  Also, check,
												report and reset singles counters. */
	char DisLin1[16];
	int TcpStatus;
	unsigned long CoinRepeat;

	if (BitRdPortI(PGDR, 7))  {																				//RUNNING?
		if (LastSinglesReport + SINGLES_PERIOD < MS_TIMER)  SinglesReport();  //handle singles counts.
		CoinRepeat = MS_TIMER + 100;
		while (MS_TIMER < CoinRepeat)  {							//KEEP CHECKING FOR COINCIDENCES FOR 100mS.
			if (BitRdPortI(PEDR, 1))  {	 														//HAVE COINCIDENCE.
				if (!CoinRptLen)  {		 								//make coincidence report packet header.
					CoinRpt[0] = 'P';
					CoinRptLen = 2;
				}

				CoinRpt[CoinRptLen++] = 'R';		 									  //coincidence report flag.

/*				BitWrPortI(PEDR, &PEDRShadow, 1, 5);	  		 			//go to any register other than 0.
				BitWrPortI(PEDR, &PEDRShadow, 0, 5);		  					  		 //ADC "A" in FPGA reg 0.
				BitWrPortI(PEDR, &PEDRShadow, 0, 6);
				BitWrPortI(PEDR, &PEDRShadow, 0, 7);
*/				#asm
				IOI LD A, (070h)		;read in PEDR
				SET 5, A					;set LSB high
				RES 6, A
				RES 7, A					;set MSB low
				IOI LD (070h), A		;write 1 to PEDR
				RES 5, A					;now set LSB low
				IOI LD (070h), A		;write 0 to PEDR
				#endasm

				CoinRpt[CoinRptLen++] = RdPortI(PGDR) & 0x0F;	 							//ADC A: MS nibble.
				CoinRpt[CoinRptLen++] = RdPortI(PFDR);				 										//LS byte.

//				BitWrPortI(PEDR, &PEDRShadow, 1, 5);				  					 //ADC "B" in FPGA reg 1.
				#asm
				IOI LD A, (070h)		;read in PEDR
				SET 5, A					;set LSB high
				IOI LD (070h), A		;write 1 to PEDR
				#endasm
				CoinRpt[CoinRptLen++] = RdPortI(PGDR) & 0x0F;								//ADC B: MS nibble.
				CoinRpt[CoinRptLen++] = RdPortI(PFDR);

//				BitWrPortI(PEDR, &PEDRShadow, 0, 5);	  		  		//offset and overrange in FPGA reg 2.
// 			BitWrPortI(PEDR, &PEDRShadow, 1, 6);
				#asm
				IOI LD A, (070h)		;read in PEDR
				RES 5, A					;set LSB low
				SET 6, A
				IOI LD (070h), A		;write 2 to PEDR
				#endasm
				CoinRpt[CoinRptLen++] = RdPortI(PFDR);						 //offset counter, 10nS per count.
				CoinRpt[CoinRptLen++] = RdPortI(PGDR) & 0x07;						  //overranges, PD order.

				CoinCount++;

				if (CoinRptLen + 7 > MTU_R)  {		 	  //FULL PACKET, SEND COINCIDENCE REPORT TO HOST.
					TcpStatus = tcp_tick(&sock);												// socket maintenance.
					CoinRpt[1] = (CoinRptLen - 2) / 7;			 //number of coincidence reports in packet.
					sock_fastwrite(&sock, CoinRpt, CoinRptLen);								 //send the report.
					CoinRptLen = 0;								  									//clear the report.
					HeartbeatLastOut = MS_TIMER;		 	 			 //no need to send a seperate heartbeat.
				}

//				BitWrPortI(PGDR, &PGDRShadow, 0, 7);	 			//acknowledge data by killing RunEnable.
//				BitWrPortI(PGDR, &PGDRShadow, 1, 7);	 									  //re-enable running.
				#asm
				IOI LD A, (048h)		;read in PGDR
				RES 7, A					;set RunEnable low
				IOI LD (048h), A		;write to PGDR
				SET 7, A					;set RunEnable back high
				IOI LD (048h), A		;write to PGDR
				#endasm
			}
		}


		if (MS_TIMER > LastDisplayUpdate + 1000)  {											  //UPDATE DISPLAY?
			LastDisplayUpdate = MS_TIMER;
			ltoa(CoinCount - CoinCountLast, DisLin1);						 //cast the long rate as a string.
			CoinCountLast = CoinCount;
//			for (Ptr = strlen(CoinCountstr) - 1 + strlen(CoinCountstr)/3; Ptr; Ptr--) {//insert commas.
//				if (Ptr % 3)  CoinCountstr[Ptr] = CoinCountstr[Ptr-1];
//				else          CoinCountstr[Ptr] = ',';
//			}
			sprintf(Msg, "Rate: %s", DisLin1);
			DisplayLumex2x16(SinglesMsg, Msg);				 //display singles rates and coincidence rate.

			if	(CoinRptLen > 0)  {													  //Send Coincidence Report(s).
				TcpStatus = tcp_tick(&sock);													// socket maintenance.
				CoinRpt[1] = (CoinRptLen - 2) / 7;				 //number of coincidence reports in packet.
				sock_fastwrite(&sock, CoinRpt, CoinRptLen);									 //send the report.
				CoinRptLen = 0;								  										//clear the report.
				HeartbeatLastOut = MS_TIMER;				  			 //no need to send a seperate heartbeat.
			}
		}
	}
}


/************************************ SinglesReport ***********************************************/
void SinglesReport(void)  {		/* Get singles count from FPGA, send to host, clear in FPGA.
			   Singles are ADC counts passing threshold per channel, whether or not a coincidence. */

	unsigned int SinglesA, SinglesB;
	char SinglesRpt[6];	  													 //buffer for singles report to host.

	LastSinglesReport = LastSinglesReport + SINGLES_PERIOD;					 //set time for next report.
	BitWrPortI(PGDR, &PGDRShadow, 0, 7);	 				  //stop running so can read singles counters.

																												//READ COUNTERS.
	BitWrPortI(PEDR, &PEDRShadow, 1, 5);		  						//Singles "A" LS chunk in FPGA reg 5.
	BitWrPortI(PEDR, &PEDRShadow, 0, 6);
	BitWrPortI(PEDR, &PEDRShadow, 1, 7);
	SinglesRpt[1] = RdPortI(PGDR) & 0x0F;														 //MS lower nibble.
	SinglesRpt[2] = RdPortI(PFDR);																			//LS byte.
	SinglesA = (unsigned int)RdPortI(PFDR) + (unsigned int)(RdPortI(PGDR) & 0x0F) * 256;
//@@printf("Reg 5: PGDR=0x%X  PFDR=0x%X\n", RdPortI(PGDR) & 0x0F, RdPortI(PFDR));

	BitWrPortI(PEDR, &PEDRShadow, 0, 5);		  						//Singles "B" LS chunk in FPGA reg 6.
	BitWrPortI(PEDR, &PEDRShadow, 1, 6);
	BitWrPortI(PEDR, &PEDRShadow, 1, 7);
	SinglesB = (unsigned int)RdPortI(PFDR) + (unsigned int)(RdPortI(PGDR) & 0x0F) * 256;

	BitWrPortI(PEDR, &PEDRShadow, 1, 5);			  						//Singles MS chunks in FPGA reg 7.
	BitWrPortI(PEDR, &PEDRShadow, 1, 6);
	BitWrPortI(PEDR, &PEDRShadow, 1, 7);
	SinglesA = SinglesA + (unsigned int)(RdPortI(PFDR) & 0x0F) * 4096;
	SinglesB = SinglesB + (unsigned int)(RdPortI(PGDR) & 0x0F) * 4096;


	SinglesRpt[0] = 'S';																				 //SEND THE REPORT.
	SinglesRpt[1] = SinglesA / 256;															//Singles "A" MS byte.
	SinglesRpt[2] = SinglesA % 256;															//Singles "A" LS byte.
	SinglesRpt[3] = SinglesB / 256;															//Singles "B" MS byte.
	SinglesRpt[4] = SinglesB % 256;															//Singles "B" LS byte.
	sock_fastwrite(&sock, SinglesRpt, 5);														 //send the report.
	sprintf(SinglesMsg,"A=%u B=%u", SinglesA, SinglesB);								 //format for display.

																						//RESET COUNTERS & RESUME RUNNING.
	BitWrPortI(PEDR, &PEDRShadow, 1, 3);					 //PE3, set calibration high (reset counters).
	BitWrPortI(PEDR, &PEDRShadow, 0, 3);											 //PE3, set calibration low.
	BitWrPortI(PGDR, &PGDRShadow, 1, 7);	 												  //re-enable running.
}


/**************************************************************************************************/